import re
from typing import TYPE_CHECKING, List, Optional, Pattern, Union

from starlite.exceptions import ImproperlyConfiguredException

if TYPE_CHECKING:
    from starlite.types import Scope, Scopes


def build_exclude_path_pattern(*, exclude: Optional[Union[str, List[str]]] = None) -> Optional[Pattern]:
    """Build single path pattern from list of patterns to opt-out from middleware processing.

    Args:
        exclude: A pattern or a list of patterns.

    Returns:
        An optional pattern to match against scope["path"] to opt-out from middleware processing.
    """
    if exclude is None:
        return None

    try:
        return re.compile("|".join(exclude)) if isinstance(exclude, list) else re.compile(exclude)
    except re.error as e:  # pragma: no cover
        raise ImproperlyConfiguredException(
            "Unable to compile exclude patterns for middleware. Please make sure you passed a valid regular expression."
        ) from e


def should_bypass_middleware(
    *,
    scope: "Scope",
    scopes: "Scopes",
    exclude_opt_key: Optional[str] = None,
    exclude_path_pattern: Optional[Pattern] = None,
) -> bool:
    """Determine weather a middleware should be bypassed.

    Args:
        scope: The ASGI scope.
        scopes: A set with the ASGI scope types that are supported by the middleware.
        exclude_opt_key: Key in ``opt`` with which a route handler can "opt-out" of a middleware.
        exclude_path_pattern: If this pattern matches scope["path"], the middleware should
            be bypassed.

    Returns:
        A boolean indicating if a middleware should be bypassed
    """
    if scope["type"] not in scopes:
        return True

    if exclude_opt_key and scope["route_handler"].opt.get(exclude_opt_key):
        return True

    if exclude_path_pattern and exclude_path_pattern.findall(
        scope["path"] if not getattr(scope.get("route_handler", {}), "is_mount", False) else scope["raw_path"].decode()
    ):
        return True
    return False



from .base import Response
from .file import FileResponse
from .redirect import RedirectResponse
from .streaming import StreamingResponse
from .template import TemplateResponse

__all__ = ["Response", "RedirectResponse", "StreamingResponse", "TemplateResponse", "FileResponse"]



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, List, Optional, Union

from pydantic import BaseConfig, BaseModel

from starlite.connection import ASGIConnection
from starlite.enums import ScopeType
from starlite.middleware.utils import (
    build_exclude_path_pattern,
    should_bypass_middleware,
)

if TYPE_CHECKING:
    from starlite.types import ASGIApp, Receive, Scope, Scopes, Send


class AuthenticationResult(BaseModel):
    """Pydantic model for authentication data."""

    user: Any
    """The user model, this can be any value corresponding to a user of the API."""
    auth: Any = None
    """The auth value, this can for example be a JWT token."""

    class Config(BaseConfig):
        arbitrary_types_allowed = True


class AbstractAuthenticationMiddleware(ABC):
    """Abstract AuthenticationMiddleware that allows users to create their own AuthenticationMiddleware by extending it
    and overriding :meth:`AbstractAuthenticationMiddleware.authenticate_request`.
    """

    def __init__(
        self,
        app: "ASGIApp",
        exclude: Optional[Union[str, List[str]]] = None,
        exclude_from_auth_key: str = "exclude_from_auth",
        scopes: Optional["Scopes"] = None,
    ) -> None:
        """Initialize ``AbstractAuthenticationMiddleware``.

        Args:
            app: An ASGIApp, this value is the next ASGI handler to call in the middleware stack.
            exclude: A pattern or list of patterns to skip in the authentication middleware.
            exclude_from_auth_key: An identifier to use on routes to disable authentication for a particular route.
            scopes: ASGI scopes processed by the authentication middleware.
        """
        self.app = app
        self.scopes = scopes or {ScopeType.HTTP, ScopeType.WEBSOCKET}
        self.exclude_opt_key = exclude_from_auth_key
        self.exclude = build_exclude_path_pattern(exclude=exclude)

    async def __call__(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        """ASGI callable.

        Args:
            scope: The ASGI connection scope.
            receive: The ASGI receive function.
            send: The ASGI send function.

        Returns:
            None
        """
        if not should_bypass_middleware(
            scope=scope,
            scopes=self.scopes,
            exclude_path_pattern=self.exclude,
            exclude_opt_key=self.exclude_opt_key,
        ):
            auth_result = await self.authenticate_request(ASGIConnection(scope))
            scope["user"] = auth_result.user
            scope["auth"] = auth_result.auth
        await self.app(scope, receive, send)

    @abstractmethod
    async def authenticate_request(self, connection: ASGIConnection) -> AuthenticationResult:  # pragma: no cover
        """Receive the http connection and return an :class:`AuthenticationResult`.

        Notes:
            - This method must be overridden by subclasses.

        Args:
            connection: An :class:`ASGIConnection <starlite.connection.ASGIConnection>` instance.

        Raises:
            NotAuthorizedException | PermissionDeniedException: if authentication fails.

        Returns:
            An instance of :class:`AuthenticationResult <starlite.middleware.authentication.AuthenticationResult>`.
        """
        raise NotImplementedError("authenticate_request must be overridden by subclasses")



from inspect import Signature

from msgspec.inspect import UNSET
from pydantic.fields import Undefined

from starlite.types import Empty

DEFAULT_ALLOWED_CORS_HEADERS = {"Accept", "Accept-Language", "Content-Language", "Content-Type"}
DEFAULT_CHUNK_SIZE = 1024 * 128  # 128KB
HTTP_RESPONSE_BODY = "http.response.body"
HTTP_RESPONSE_START = "http.response.start"
ONE_MEGABYTE = 1024 * 1024
REDIRECT_STATUS_CODES = {301, 302, 303, 307, 308}
RESERVED_KWARGS = {"state", "headers", "cookies", "request", "socket", "data", "query", "scope", "body"}
SCOPE_STATE_DEPENDENCY_CACHE = "dependency_cache"
SCOPE_STATE_NAMESPACE = "__starlite__"
SCOPE_STATE_RESPONSE_COMPRESSED = "response_compressed"
UNDEFINED_SENTINELS = {Undefined, Signature.empty, UNSET, Empty, Ellipsis}
SKIP_VALIDATION_NAMES = {"request", "socket", "scope", "receive", "send"}



from .config import SQLAlchemyConfig, SQLAlchemyEngineConfig, SQLAlchemySessionConfig
from .plugin import SQLAlchemyPlugin

__all__ = ("SQLAlchemyPlugin", "SQLAlchemyConfig", "SQLAlchemyEngineConfig", "SQLAlchemySessionConfig")



from inspect import isclass
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
)

from pydantic import BaseConfig, DirectoryPath, PrivateAttr, root_validator
from pydantic.generics import GenericModel

from starlite.template import TemplateEngineProtocol

T = TypeVar("T", bound=TemplateEngineProtocol)


class TemplateConfig(Generic[T], GenericModel):
    """Configuration for Templating.

    To enable templating, pass an instance of this class to the :class:`Starlite <starlite.app.Starlite>` constructor using the
    'template_config' key.
    """

    class Config(BaseConfig):
        arbitrary_types_allowed = True

    directory: Optional[Union[DirectoryPath, List[DirectoryPath]]] = None
    """A directory or list of directories from which to serve templates."""
    engine: Union[Type[T], T]
    """A template engine adhering to the :class:`TemplateEngineProtocol <starlite.template.base.TemplateEngineProtocol>`."""
    engine_callback: Optional[Callable[[T], None]] = None
    """A callback function that allows modifying the instantiated templating protocol."""
    _engine_instance: T = PrivateAttr()

    @root_validator
    def validate_config(cls, values: Dict[str, Any]) -> Dict[str, Any]:  # pylint: disable=E0213
        """Ensure that directory is set if engine is a class.

        Args:
            values: The dictionary of values to validate

        Returns:
            The validated dictionary of values.
        """
        if isclass(values.get("engine")) and not values.get("directory"):
            raise ValueError("directory is a required kwarg when passing a template engine class")
        return values

    def to_engine(self) -> T:
        """Instantiate the template engine."""
        template_engine = cast("T", self.engine(self.directory) if isclass(self.engine) else self.engine)
        if callable(self.engine_callback):
            self.engine_callback(template_engine)  # pylint: disable=E1102
        self._engine_instance = template_engine
        return template_engine

    @property
    def engine_instance(self) -> T:
        """Return the template engine instance."""
        if not hasattr(self, "_engine_instance"):
            return self.to_engine()
        return self._engine_instance



from functools import partial
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Tuple, Union

from starlite.exceptions import MissingDependencyException, TemplateNotFoundException
from starlite.template.base import (
    TemplateEngineProtocol,
    TemplateProtocol,
    csrf_token,
    url_for,
    url_for_static_asset,
)

try:
    from mako.exceptions import TemplateLookupException as MakoTemplateNotFound
    from mako.lookup import TemplateLookup

except ImportError as e:
    raise MissingDependencyException("mako is not installed") from e

if TYPE_CHECKING:
    from mako.template import Template as _MakoTemplate
    from pydantic import DirectoryPath


class MakoTemplate(TemplateProtocol):
    """Mako template, implementing ``TemplateProtocol``"""

    def __init__(
        self, template: "_MakoTemplate", template_callables: List[Tuple[str, Callable[[Dict[str, Any]], Any]]]
    ):
        """Initialize a template.

        Args:
            template: Base ``MakoTemplate`` used by the underlying mako-engine
            template_callables: List of callables passed to the template
        """
        super().__init__()
        self.template = template
        self.template_callables = template_callables

    def render(self, *args: Any, **kwargs: Any) -> str:
        """Render a template.

        Args:
            args: Positional arguments passed to the engines ``render`` function
            kwargs: Keyword arguments passed to the engines ``render`` function

        Returns:
            Rendered template as a string
        """
        for callable_key, template_callable in self.template_callables:
            kwargs_copy = {**kwargs}
            kwargs[callable_key] = partial(template_callable, kwargs_copy)

        return str(self.template.render(*args, **kwargs))


class MakoTemplateEngine(TemplateEngineProtocol[MakoTemplate]):
    """Mako based TemplateEngine."""

    def __init__(self, directory: Union["DirectoryPath", List["DirectoryPath"]]) -> None:
        """Initialize template engine.

        Args:
            directory: Direct path or list of directory paths from which to serve templates.
        """
        super().__init__(directory=directory)
        self.engine = TemplateLookup(directories=directory if isinstance(directory, (list, tuple)) else [directory])
        self._template_callables: List[Tuple[str, Callable[[Dict[str, Any]], Any]]] = []
        self.register_template_callable(key="url_for_static_asset", template_callable=url_for_static_asset)  # type: ignore
        self.register_template_callable(key="csrf_token", template_callable=csrf_token)  # type: ignore
        self.register_template_callable(key="url_for", template_callable=url_for)  # type: ignore

    def get_template(self, template_name: str) -> MakoTemplate:
        """Retrieve a template by matching its name (dotted path) with files in the directory or directories provided.
        Args:
            template_name: A dotted path

        Returns:
            MakoTemplate instance

        Raises:
            :class:`TemplateNotFoundException <starlite.exceptions.TemplateNotFoundException>`: if no template is found.
        """
        try:
            return MakoTemplate(
                template=self.engine.get_template(template_name), template_callables=self._template_callables
            )
        except MakoTemplateNotFound as exc:
            raise TemplateNotFoundException(template_name=template_name) from exc

    def register_template_callable(self, key: str, template_callable: Callable[[Dict[str, Any]], Any]) -> None:
        """Register a callable on the template engine.

        Args:
            key: The callable key, i.e. the value to use inside the template to call the callable.
            template_callable: A callable to register.

        Returns:
            None
        """
        self._template_callables.append((key, template_callable))



"""Includes code adapted from https://github.com/encode/starlette/blob/master/starlette/status.py.

Copyright © 2018, [Encode OSS Ltd](https://www.encode.io/).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from typing import Literal

# HTTP Status Codes

HTTP_100_CONTINUE: Literal[100] = 100
HTTP_101_SWITCHING_PROTOCOLS: Literal[101] = 101
HTTP_102_PROCESSING: Literal[102] = 102
HTTP_103_EARLY_HINTS: Literal[103] = 103
HTTP_200_OK: Literal[200] = 200
HTTP_201_CREATED: Literal[201] = 201
HTTP_202_ACCEPTED: Literal[202] = 202
HTTP_203_NON_AUTHORITATIVE_INFORMATION: Literal[203] = 203
HTTP_204_NO_CONTENT: Literal[204] = 204
HTTP_205_RESET_CONTENT: Literal[205] = 205
HTTP_206_PARTIAL_CONTENT: Literal[206] = 206
HTTP_207_MULTI_STATUS: Literal[207] = 207
HTTP_208_ALREADY_REPORTED: Literal[208] = 208
HTTP_226_IM_USED: Literal[226] = 226
HTTP_300_MULTIPLE_CHOICES: Literal[300] = 300
HTTP_301_MOVED_PERMANENTLY: Literal[301] = 301
HTTP_302_FOUND: Literal[302] = 302
HTTP_303_SEE_OTHER: Literal[303] = 303
HTTP_304_NOT_MODIFIED: Literal[304] = 304
HTTP_305_USE_PROXY: Literal[305] = 305
HTTP_306_RESERVED: Literal[306] = 306
HTTP_307_TEMPORARY_REDIRECT: Literal[307] = 307
HTTP_308_PERMANENT_REDIRECT: Literal[308] = 308
HTTP_400_BAD_REQUEST: Literal[400] = 400
HTTP_401_UNAUTHORIZED: Literal[401] = 401
HTTP_402_PAYMENT_REQUIRED: Literal[402] = 402
HTTP_403_FORBIDDEN: Literal[403] = 403
HTTP_404_NOT_FOUND: Literal[404] = 404
HTTP_405_METHOD_NOT_ALLOWED: Literal[405] = 405
HTTP_406_NOT_ACCEPTABLE: Literal[406] = 406
HTTP_407_PROXY_AUTHENTICATION_REQUIRED: Literal[407] = 407
HTTP_408_REQUEST_TIMEOUT: Literal[408] = 408
HTTP_409_CONFLICT: Literal[409] = 409
HTTP_410_GONE: Literal[410] = 410
HTTP_411_LENGTH_REQUIRED: Literal[411] = 411
HTTP_412_PRECONDITION_FAILED: Literal[412] = 412
HTTP_413_REQUEST_ENTITY_TOO_LARGE: Literal[413] = 413
HTTP_414_REQUEST_URI_TOO_LONG: Literal[414] = 414
HTTP_415_UNSUPPORTED_MEDIA_TYPE: Literal[415] = 415
HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE: Literal[416] = 416
HTTP_417_EXPECTATION_FAILED: Literal[417] = 417
HTTP_418_IM_A_TEAPOT: Literal[418] = 418
HTTP_421_MISDIRECTED_REQUEST: Literal[421] = 421
HTTP_422_UNPROCESSABLE_ENTITY: Literal[422] = 422
HTTP_423_LOCKED: Literal[423] = 423
HTTP_424_FAILED_DEPENDENCY: Literal[424] = 424
HTTP_425_TOO_EARLY: Literal[425] = 425
HTTP_426_UPGRADE_REQUIRED: Literal[426] = 426
HTTP_428_PRECONDITION_REQUIRED: Literal[428] = 428
HTTP_429_TOO_MANY_REQUESTS: Literal[429] = 429
HTTP_431_REQUEST_HEADER_FIELDS_TOO_LARGE: Literal[431] = 431
HTTP_451_UNAVAILABLE_FOR_LEGAL_REASONS: Literal[451] = 451
HTTP_500_INTERNAL_SERVER_ERROR: Literal[500] = 500
HTTP_501_NOT_IMPLEMENTED: Literal[501] = 501
HTTP_502_BAD_GATEWAY: Literal[502] = 502
HTTP_503_SERVICE_UNAVAILABLE: Literal[503] = 503
HTTP_504_GATEWAY_TIMEOUT: Literal[504] = 504
HTTP_505_HTTP_VERSION_NOT_SUPPORTED: Literal[505] = 505
HTTP_506_VARIANT_ALSO_NEGOTIATES: Literal[506] = 506
HTTP_507_INSUFFICIENT_STORAGE: Literal[507] = 507
HTTP_508_LOOP_DETECTED: Literal[508] = 508
HTTP_510_NOT_EXTENDED: Literal[510] = 510
HTTP_511_NETWORK_AUTHENTICATION_REQUIRED: Literal[511] = 511

# Websocket Codes
WS_1000_NORMAL_CLOSURE: Literal[1000] = 1000
WS_1001_GOING_AWAY: Literal[1001] = 1001
WS_1002_PROTOCOL_ERROR: Literal[1002] = 1002
WS_1003_UNSUPPORTED_DATA: Literal[1003] = 1003
WS_1005_NO_STATUS_RECEIVED: Literal[1005] = 1005
WS_1006_ABNORMAL_CLOSURE: Literal[1006] = 1006
WS_1007_INVALID_FRAME_PAYLOAD_DATA: Literal[1007] = 1007
WS_1008_POLICY_VIOLATION: Literal[1008] = 1008
WS_1009_MESSAGE_TOO_BIG: Literal[1009] = 1009
WS_1010_MANDATORY_EXT: Literal[1010] = 1010
WS_1011_INTERNAL_ERROR: Literal[1011] = 1011
WS_1012_SERVICE_RESTART: Literal[1012] = 1012
WS_1013_TRY_AGAIN_LATER: Literal[1013] = 1013
WS_1014_BAD_GATEWAY: Literal[1014] = 1014
WS_1015_TLS_HANDSHAKE: Literal[1015] = 1015


__all__ = [
    "HTTP_100_CONTINUE",
    "HTTP_101_SWITCHING_PROTOCOLS",
    "HTTP_102_PROCESSING",
    "HTTP_103_EARLY_HINTS",
    "HTTP_200_OK",
    "HTTP_201_CREATED",
    "HTTP_202_ACCEPTED",
    "HTTP_203_NON_AUTHORITATIVE_INFORMATION",
    "HTTP_204_NO_CONTENT",
    "HTTP_205_RESET_CONTENT",
    "HTTP_206_PARTIAL_CONTENT",
    "HTTP_207_MULTI_STATUS",
    "HTTP_208_ALREADY_REPORTED",
    "HTTP_226_IM_USED",
    "HTTP_300_MULTIPLE_CHOICES",
    "HTTP_301_MOVED_PERMANENTLY",
    "HTTP_302_FOUND",
    "HTTP_303_SEE_OTHER",
    "HTTP_304_NOT_MODIFIED",
    "HTTP_305_USE_PROXY",
    "HTTP_306_RESERVED",
    "HTTP_307_TEMPORARY_REDIRECT",
    "HTTP_308_PERMANENT_REDIRECT",
    "HTTP_400_BAD_REQUEST",
    "HTTP_401_UNAUTHORIZED",
    "HTTP_402_PAYMENT_REQUIRED",
    "HTTP_403_FORBIDDEN",
    "HTTP_404_NOT_FOUND",
    "HTTP_405_METHOD_NOT_ALLOWED",
    "HTTP_406_NOT_ACCEPTABLE",
    "HTTP_407_PROXY_AUTHENTICATION_REQUIRED",
    "HTTP_408_REQUEST_TIMEOUT",
    "HTTP_409_CONFLICT",
    "HTTP_410_GONE",
    "HTTP_411_LENGTH_REQUIRED",
    "HTTP_412_PRECONDITION_FAILED",
    "HTTP_413_REQUEST_ENTITY_TOO_LARGE",
    "HTTP_414_REQUEST_URI_TOO_LONG",
    "HTTP_415_UNSUPPORTED_MEDIA_TYPE",
    "HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE",
    "HTTP_417_EXPECTATION_FAILED",
    "HTTP_418_IM_A_TEAPOT",
    "HTTP_421_MISDIRECTED_REQUEST",
    "HTTP_422_UNPROCESSABLE_ENTITY",
    "HTTP_423_LOCKED",
    "HTTP_424_FAILED_DEPENDENCY",
    "HTTP_425_TOO_EARLY",
    "HTTP_426_UPGRADE_REQUIRED",
    "HTTP_428_PRECONDITION_REQUIRED",
    "HTTP_429_TOO_MANY_REQUESTS",
    "HTTP_431_REQUEST_HEADER_FIELDS_TOO_LARGE",
    "HTTP_451_UNAVAILABLE_FOR_LEGAL_REASONS",
    "HTTP_500_INTERNAL_SERVER_ERROR",
    "HTTP_501_NOT_IMPLEMENTED",
    "HTTP_502_BAD_GATEWAY",
    "HTTP_503_SERVICE_UNAVAILABLE",
    "HTTP_504_GATEWAY_TIMEOUT",
    "HTTP_505_HTTP_VERSION_NOT_SUPPORTED",
    "HTTP_506_VARIANT_ALSO_NEGOTIATES",
    "HTTP_507_INSUFFICIENT_STORAGE",
    "HTTP_508_LOOP_DETECTED",
    "HTTP_510_NOT_EXTENDED",
    "HTTP_511_NETWORK_AUTHENTICATION_REQUIRED",
    "WS_1000_NORMAL_CLOSURE",
    "WS_1001_GOING_AWAY",
    "WS_1002_PROTOCOL_ERROR",
    "WS_1003_UNSUPPORTED_DATA",
    "WS_1005_NO_STATUS_RECEIVED",
    "WS_1006_ABNORMAL_CLOSURE",
    "WS_1007_INVALID_FRAME_PAYLOAD_DATA",
    "WS_1008_POLICY_VIOLATION",
    "WS_1009_MESSAGE_TOO_BIG",
    "WS_1010_MANDATORY_EXT",
    "WS_1011_INTERNAL_ERROR",
    "WS_1012_SERVICE_RESTART",
    "WS_1013_TRY_AGAIN_LATER",
    "WS_1014_BAD_GATEWAY",
    "WS_1015_TLS_HANDSHAKE",
]



from typing import Any, Callable, Iterable

from anyio import create_task_group
from typing_extensions import ParamSpec

from starlite.utils.sync import AsyncCallable

P = ParamSpec("P")


class BackgroundTask:
    """A container for a 'background' task function.

    Background tasks are called once a Response finishes.
    """

    __slots__ = ("fn", "args", "kwargs")

    def __init__(self, fn: Callable[P, Any], *args: P.args, **kwargs: P.kwargs) -> None:
        """Initialize ``BackgroundTask``.

        Args:
            fn: A sync or async function to call as the background task.
            *args: Args to pass to the func.
            **kwargs: Kwargs to pass to the func
        """
        self.fn = AsyncCallable(fn)
        self.args = args
        self.kwargs = kwargs

    async def __call__(self) -> None:
        """Call the wrapped function with the passed in arguments.

        Returns:
            None
        """
        await self.fn(*self.args, **self.kwargs)


class BackgroundTasks:
    """A container for multiple 'background' task functions.

    Background tasks are called once a Response finishes.
    """

    __slots__ = ("tasks", "run_in_task_group")

    def __init__(self, tasks: Iterable[BackgroundTask], run_in_task_group: bool = False) -> None:
        """Initialize ``BackgroundTasks``.

        Args:
            tasks: An iterable of :class:`BackgroundTask <starlite.datastructures.BackgroundTask>` instances.
            run_in_task_group: If you set this value to ``True`` than the tasks will run concurrently, using
                an [anyio.task_group](https://anyio.readthedocs.io/en/stable/tasks.html). Note: this will
                not preserve execution order.
        """
        self.tasks = tasks
        self.run_in_task_group = run_in_task_group

    async def __call__(self) -> None:
        """Call the wrapped background tasks.

        Returns:
            None
        """
        if self.run_in_task_group:
            async with create_task_group() as task_group:
                for task in self.tasks:
                    task_group.start_soon(task)
        else:
            for task in self.tasks:
                await task()



from typing import TYPE_CHECKING, Any, Dict, Type, cast

from pydantic import BaseConfig, BaseModel, create_model, create_model_from_typeddict
from pydantic_factories.utils import create_model_from_dataclass

if TYPE_CHECKING:
    from pydantic.fields import ModelField

    from starlite.types.builtin_types import DataclassClassOrInstance, TypedDictClass


class Config(BaseConfig):
    """Base config for models."""

    arbitrary_types_allowed = True


def create_parsed_model_field(value: Type[Any]) -> "ModelField":
    """Create a pydantic model with the passed in value as its sole field, and return the parsed field."""
    model = create_model("temp", __config__=Config, **{"value": (value, ... if not repr(value).startswith("typing.Optional") else None)})  # type: ignore
    return cast("BaseModel", model).__fields__["value"]


_type_model_map: Dict[Type[Any], Type[BaseModel]] = {}


def convert_dataclass_to_model(dataclass_or_instance: "DataclassClassOrInstance") -> Type[BaseModel]:
    """Convert a dataclass or dataclass instance to a pydantic model and memoize the result."""
    if not isinstance(dataclass_or_instance, type):
        dataclass = type(dataclass_or_instance)
    else:
        dataclass = dataclass_or_instance

    existing = _type_model_map.get(dataclass)
    if not existing:
        _type_model_map[dataclass] = existing = create_model_from_dataclass(dataclass)
    return existing


def convert_typeddict_to_model(typeddict: "TypedDictClass") -> Type[BaseModel]:
    """Convert a :class:`TypedDict <typing.TypedDict>` to a pydantic model and memoize the result."""
    existing = _type_model_map.get(typeddict)
    if not existing:
        _type_model_map[typeddict] = existing = create_model_from_typeddict(typeddict)
    return existing



from typing import Awaitable, List, Literal, TypeVar, Union

T = TypeVar("T")

SyncOrAsyncUnion = Union[T, Awaitable[T]]
"""Types 'T' as a union of T and awaitable T."""
SingleOrList = Union[T, List[T]]
"""Types 'T' as a single value or a list T."""
AnyIOBackend = Literal["asyncio", "trio"]
"""Anyio backend names."""



from starlite.contrib.jwt.jwt_auth import (
    JWTAuth,
    JWTCookieAuth,
    OAuth2Login,
    OAuth2PasswordBearerAuth,
)
from starlite.contrib.jwt.jwt_token import Token
from starlite.contrib.jwt.middleware import (
    JWTAuthenticationMiddleware,
    JWTCookieAuthenticationMiddleware,
)

__all__ = (
    "JWTAuth",
    "JWTAuthenticationMiddleware",
    "JWTCookieAuth",
    "JWTCookieAuthenticationMiddleware",
    "OAuth2Login",
    "OAuth2PasswordBearerAuth",
    "Token",
)



import inspect

from click import command, option
from rich.tree import Tree

from starlite import HTTPRoute, Starlite, WebSocketRoute
from starlite.cli.utils import StarliteCLIException, StarliteEnv, console, show_app_info
from starlite.utils.helpers import unwrap_partial


@command(name="info")
def info_command(app: Starlite) -> None:
    """Show information about the detected Starlite app."""

    show_app_info(app)


@command(name="run")
@option("-r", "--reload", help="Reload server on changes", default=False, is_flag=True)
@option("-p", "--port", help="Serve under this port", type=int, default=8000, show_default=True)
@option("--host", help="Server under this host", default="127.0.0.1", show_default=True)
@option("--debug", help="Run app in debug mode", is_flag=True)
def run_command(
    reload: bool,
    port: int,
    host: str,
    debug: bool,
    env: StarliteEnv,
    app: Starlite,
) -> None:
    """Run a Starlite app.

    The app can be either passed as a module path in the form of <module name>.<submodule>:<app instance or factory>,
    set as an environment variable STARLITE_APP with the same format or automatically discovered from one of these
    canonical paths: app.py, asgi.py, application.py or app/__init__.py. When autodiscovering application factories,
    functions with the name ``create_app`` are considered, or functions that are annotated as returning a ``Starlite``
    instance.
    """

    try:
        import uvicorn
    except ImportError:
        raise StarliteCLIException("Uvicorn needs to be installed to run an app")  # pylint: disable=W0707

    if debug or env.debug:
        app.debug = True

    show_app_info(app)

    console.rule("[yellow]Starting server process", align="left")

    uvicorn.run(
        env.app_path,
        reload=env.reload or reload,
        host=env.host or host,
        port=env.port or port,
        factory=env.is_app_factory,
    )


@command(name="routes")
def routes_command(app: Starlite) -> None:  # pragma: no cover
    """Display information about the application's routes."""

    tree = Tree("", hide_root=True)

    for route in sorted(app.routes, key=lambda r: r.path):
        if isinstance(route, HTTPRoute):
            branch = tree.add(f"[green]{route.path}[/green] (HTTP)")
            for handler in route.route_handlers:
                handler_info = [
                    f"[blue]{handler.name or handler.handler_name}[/blue]",
                ]

                if inspect.iscoroutinefunction(unwrap_partial(handler.fn.value)):
                    handler_info.append("[magenta]async[/magenta]")
                else:
                    handler_info.append("[yellow]sync[/yellow]")

                handler_info.append(f'[cyan]{", ".join(sorted(handler.http_methods))}[/cyan]')

                if len(handler.paths) > 1:
                    for path in handler.paths:
                        branch.add(" ".join([f"[green]{path}[green]", *handler_info]))
                else:
                    branch.add(" ".join(handler_info))

        else:
            if isinstance(route, WebSocketRoute):
                route_type = "WS"
            else:
                route_type = "ASGI"
            branch = tree.add(f"[green]{route.path}[/green] ({route_type})")
            branch.add(f"[blue]{route.route_handler.name or route.route_handler.handler_name}[/blue]")

    console.print(tree)



from typing import List, Literal, Optional, Set, Union

from pydantic import BaseModel

from starlite.types import Method


class CSRFConfig(BaseModel):
    """Configuration for CSRF (Cross Site Request Forgery) protection.

    To enable CSRF protection, pass an instance of this class to the :class:`Starlite <starlite.app.Starlite>` constructor using
    the 'csrf_config' key.
    """

    secret: str
    """A string that is used to create an HMAC to sign the CSRF token."""
    cookie_name: str = "csrftoken"
    """The CSRF cookie name."""
    cookie_path: str = "/"
    """The CSRF cookie path."""
    header_name: str = "x-csrftoken"
    """The header that will be expected in each request."""
    cookie_secure: bool = False
    """A boolean value indicating whether to set the ``Secure`` attribute on the cookie."""
    cookie_httponly: bool = False
    """A boolean value indicating whether to set the ``HttpOnly`` attribute on the cookie."""
    cookie_samesite: Literal["lax", "strict", "none"] = "lax"
    """The value to set in the ``SameSite`` attribute of the cookie."""
    cookie_domain: Optional[str] = None
    """Specifies which hosts can receive the cookie."""
    safe_methods: Set[Method] = {"GET", "HEAD"}
    """A set of "safe methods" that can set the cookie."""
    exclude: Optional[Union[str, List[str]]] = None
    """A pattern or list of patterns to skip in the CSRF middleware."""
    exclude_from_csrf_key: str = "exclude_from_csrf"
    """An identifier to use on routes to disable CSRF for a particular route."""



from typing import TYPE_CHECKING, Dict, List, Tuple

from pydantic_openapi_schema.v3_1_0.parameter import Parameter

from starlite.constants import RESERVED_KWARGS
from starlite.enums import ParamType
from starlite.exceptions import ImproperlyConfiguredException
from starlite.openapi.schema import create_schema
from starlite.params import DependencyKwarg, ParameterKwarg
from starlite.signature.models import SignatureField
from starlite.types import Empty

if TYPE_CHECKING:
    from pydantic_openapi_schema.v3_1_0.schema import Schema

    from starlite.handlers import BaseRouteHandler
    from starlite.types import Dependencies
    from starlite.types.internal_types import PathParameterDefinition


def create_path_parameter_schema(
    path_parameter: "PathParameterDefinition", field: "SignatureField", generate_examples: bool
) -> "Schema":
    """Create a path parameter from the given path_param definition."""
    return create_schema(
        field=SignatureField(
            name=field.name,
            field_type=path_parameter.type,
            children=None,
            extra=field.extra,
            kwarg_model=field.kwarg_model,
            default_value=field.default_value,
        ),
        generate_examples=generate_examples,
        plugins=[],
    )


class ParameterCollection:
    """Facilitates conditional deduplication of parameters.

    If multiple parameters with the same name are produced for a handler, the condition is ignored if the two
    ``Parameter`` instances are the same (the first is retained and any duplicates are ignored). If the ``Parameter``
    instances are not the same, an exception is raised.
    """

    def __init__(self, route_handler: "BaseRouteHandler") -> None:
        """Initialize ``ParameterCollection``.

        Args:
            route_handler: Associated route handler
        """
        self.route_handler = route_handler
        self._parameters: Dict[str, Parameter] = {}

    def add(self, parameter: Parameter) -> None:
        """Add a ``Parameter`` to the collection.

        If an existing parameter with the same name and type already exists, the
        parameter is ignored.

        If an existing parameter with the same name but different type exists, raises
        ``ImproperlyConfiguredException``.
        """
        if parameter.name not in self._parameters:
            self._parameters[parameter.name] = parameter
            return

        pre_existing = self._parameters[parameter.name]
        if parameter == pre_existing:
            return

        raise ImproperlyConfiguredException(
            f"OpenAPI schema generation for handler `{self.route_handler}` detected multiple parameters named "
            f"'{parameter.name}' with different types."
        )

    def list(self) -> List[Parameter]:
        """Return a list of all ``Parameter``'s in the collection."""
        return list(self._parameters.values())


def create_parameter(
    signature_field: "SignatureField",
    parameter_name: str,
    path_parameters: Tuple["PathParameterDefinition", ...],
    generate_examples: bool,
) -> Parameter:
    """Create an OpenAPI Parameter instance."""
    schema = None
    kwargs_model = signature_field.kwarg_model if isinstance(signature_field.kwarg_model, ParameterKwarg) else None

    if any(path_param.name == parameter_name for path_param in path_parameters):
        param_in = ParamType.PATH
        is_required = True
        path_parameter = [p for p in path_parameters if parameter_name in p.name][0]
        schema = create_path_parameter_schema(
            path_parameter=path_parameter,
            field=signature_field,
            generate_examples=generate_examples,
        )

    elif kwargs_model and kwargs_model.header:
        parameter_name = kwargs_model.header
        param_in = ParamType.HEADER
        is_required = signature_field.is_required
    elif kwargs_model and kwargs_model.cookie:
        parameter_name = kwargs_model.cookie
        param_in = ParamType.COOKIE
        is_required = signature_field.is_required
    else:
        is_required = signature_field.is_required
        param_in = ParamType.QUERY
        parameter_name = kwargs_model.query if kwargs_model and kwargs_model.query else parameter_name

    if not schema:
        schema = create_schema(field=signature_field, generate_examples=generate_examples, plugins=[])

    return Parameter(
        name=parameter_name,
        param_in=param_in,
        required=is_required,
        param_schema=schema,
        description=schema.description,
    )


def get_recursive_handler_parameters(
    field_name: str,
    signature_field: "SignatureField",
    dependencies: "Dependencies",
    route_handler: "BaseRouteHandler",
    path_parameters: Tuple["PathParameterDefinition", ...],
    generate_examples: bool,
) -> List[Parameter]:
    """Create and return parameters for a handler.

    If the provided field is not a dependency, a normal parameter is created and returned as a list, otherwise
    `create_parameter_for_handler()` is called to generate parameters for the dependency.
    """

    if field_name not in dependencies:
        return [
            create_parameter(
                signature_field=signature_field,
                parameter_name=field_name,
                path_parameters=path_parameters,
                generate_examples=generate_examples,
            )
        ]
    dependency_fields = dependencies[field_name].signature_model.fields()
    return create_parameter_for_handler(route_handler, dependency_fields, path_parameters, generate_examples)


def get_layered_parameter(
    field_name: str,
    signature_field: "SignatureField",
    layered_parameters: Dict[str, "SignatureField"],
    path_parameters: Tuple["PathParameterDefinition", ...],
    generate_examples: bool,
) -> Parameter:
    """Create a layered parameter for a given signature model field.

    Layer info is extracted from the provided ``layered_parameters`` dict and set as the field's ``field_info`` attribute.
    """
    layer_field = layered_parameters[field_name]

    field = signature_field if signature_field.is_parameter_field else layer_field
    default_value = signature_field.default_value if not signature_field.is_empty else layer_field.default_value
    field_type = signature_field.field_type if signature_field is not Empty else layer_field.field_type  # type: ignore

    parameter_name = field_name
    if isinstance(field.kwarg_model, ParameterKwarg):
        parameter_name = field.kwarg_model.query or field.kwarg_model.header or field.kwarg_model.cookie or field_name

    return create_parameter(
        signature_field=SignatureField.create(
            default_value=default_value,
            kwarg_model=field.kwarg_model,
            name=field_name,
            field_type=field_type,
            extra=field.extra,
            children=field.children,
        ),
        parameter_name=parameter_name,
        path_parameters=path_parameters,
        generate_examples=generate_examples,
    )


def create_parameter_for_handler(
    route_handler: "BaseRouteHandler",
    handler_fields: Dict[str, "SignatureField"],
    path_parameters: Tuple["PathParameterDefinition", ...],
    generate_examples: bool,
) -> List[Parameter]:
    """Create a list of path/query/header Parameter models for the given PathHandler."""
    parameters = ParameterCollection(route_handler=route_handler)
    dependencies = route_handler.resolve_dependencies()

    layered_parameters = route_handler.resolve_layered_parameters()

    unique_handler_fields = tuple(
        (k, v) for k, v in handler_fields.items() if k not in RESERVED_KWARGS and k not in layered_parameters
    )
    unique_layered_fields = tuple(
        (k, v) for k, v in layered_parameters.items() if k not in RESERVED_KWARGS and k not in handler_fields
    )
    intersection_fields = tuple(
        (k, v) for k, v in handler_fields.items() if k not in RESERVED_KWARGS and k in layered_parameters
    )

    for field_name, signature_field in unique_handler_fields:
        if isinstance(signature_field.kwarg_model, DependencyKwarg) and field_name not in dependencies:
            # never document explicit dependencies
            continue

        for parameter in get_recursive_handler_parameters(
            field_name=field_name,
            signature_field=signature_field,
            dependencies=dependencies,
            route_handler=route_handler,
            path_parameters=path_parameters,
            generate_examples=generate_examples,
        ):
            parameters.add(parameter)

    for field_name, signature_field in unique_layered_fields:
        parameters.add(
            create_parameter(
                signature_field=signature_field,
                parameter_name=field_name,
                path_parameters=path_parameters,
                generate_examples=generate_examples,
            )
        )

    for field_name, signature_field in intersection_fields:
        parameters.add(
            get_layered_parameter(
                field_name=field_name,
                signature_field=signature_field,
                layered_parameters=layered_parameters,
                path_parameters=path_parameters,
                generate_examples=generate_examples,
            )
        )

    return parameters.list()



import re
from typing import TYPE_CHECKING

from starlite.exceptions import ImproperlyConfiguredException
from starlite.openapi.enums import OpenAPIType

if TYPE_CHECKING:
    from starlite.signature.models import SignatureField

CAPITAL_LETTERS_PATTERN = re.compile(r"(?=[A-Z])")


def pascal_case_to_text(string: str) -> str:
    """Given a 'PascalCased' string, return its split form- 'Pascal Cased'."""
    return " ".join(re.split(CAPITAL_LETTERS_PATTERN, string)).strip()


def get_openapi_type_for_complex_type(field: "SignatureField") -> "OpenAPIType":
    """We are dealing with complex types in this case.

    The problem here is that the Python typing system is too crude to define OpenAPI objects properly.
    """
    if field.is_mapping:
        return OpenAPIType.OBJECT
    if field.is_non_string_sequence or field.is_non_string_iterable:
        return OpenAPIType.ARRAY

    raise ImproperlyConfiguredException(  # pragma: no cover
        f"Parameter '{field.name}' with type '{field.field_type}' could not be mapped to an Open API type. "
        f"This can occur if a user-defined generic type is resolved as a parameter. If '{field.name}' should "
        "not be documented as a parameter, annotate it using the `Dependency` function, e.g., "
        f"`{field.name}: ... = Dependency(...)`."
    )



from typing import TYPE_CHECKING, List, cast

if TYPE_CHECKING:
    from typing import Union

    from starlite.routes import ASGIRoute, HTTPRoute, WebSocketRoute
    from starlite.routes.base import BaseRoute
    from starlite.types import ASGIApp, ExceptionHandlersMap, RouteHandlerType


def wrap_in_exception_handler(debug: bool, app: "ASGIApp", exception_handlers: "ExceptionHandlersMap") -> "ASGIApp":
    """Wrap the given ASGIApp in an instance of ExceptionHandlerMiddleware.

    Args:
        debug: Dictates whether exceptions are raised in debug mode.
        app: The ASGI app that is being wrapped.
        exception_handlers: A mapping of exceptions to handler functions.

    Returns:
        A wrapped ASGIApp.
    """
    from starlite.middleware.exceptions import ExceptionHandlerMiddleware

    return ExceptionHandlerMiddleware(app=app, exception_handlers=exception_handlers, debug=debug)


def get_route_handlers(route: "BaseRoute") -> List["RouteHandlerType"]:
    """Retrieve handler(s) as a list for given route.

    Args:
        route: The route from which the route handlers are extracted.

    Returns:
        The route handlers defined on the route.
    """
    route_handlers: List["RouteHandlerType"] = []
    if hasattr(route, "route_handlers"):
        route_handlers.extend(cast("HTTPRoute", route).route_handlers)
    else:
        route_handlers.append(cast("Union[WebSocketRoute, ASGIRoute]", route).route_handler)

    return route_handlers



from typing import TYPE_CHECKING, Any, Dict, Type

from starlite.exceptions import MissingDependencyException
from starlite.plugins.base import PluginProtocol

try:
    from piccolo.table import Table, TableMetaclass
    from piccolo.utils.pydantic import create_pydantic_model
except ImportError as e:
    raise MissingDependencyException("piccolo orm is not installed") from e

if TYPE_CHECKING:
    from pydantic import BaseModel
    from typing_extensions import TypeGuard


class PiccoloORMPlugin(PluginProtocol[Table]):
    """Support (de)serialization and OpenAPI generation for Piccolo ORM types."""

    _models_map: Dict[Type[Table], Type["BaseModel"]] = {}
    _data_models_map: Dict[Type[Table], Type["BaseModel"]] = {}

    def to_pydantic_model_class(self, model_class: Type[Table], **kwargs: Any) -> Type["BaseModel"]:
        """Given a piccolo model_class instance, convert it to a subclass of the piccolo "BaseModel".

        Since incoming request body's cannot and should not include values for
        related fields, pk fields and read only fields in piccolo-orm, we generate two different kinds of pydantic models here:
        - the first is a regular pydantic model, and the other is for the "data" kwarg only, which is further sanitized.

        This function uses memoization to ensure we don't recompute unnecessarily.
        """
        parameter_name = kwargs.get("parameter_name")
        if parameter_name == "data":
            if model_class not in self._data_models_map:
                self._data_models_map[model_class] = create_pydantic_model(
                    table=model_class, model_name=f"{model_class.__name__}RequestBody"
                )
            return self._data_models_map[model_class]
        if model_class not in self._models_map:
            self._models_map[model_class] = create_pydantic_model(
                table=model_class, model_name=model_class.__name__, nested=True, include_default_columns=True
            )
        return self._models_map[model_class]

    @staticmethod
    def is_plugin_supported_type(value: Any) -> "TypeGuard[Table]":
        """Given a value of indeterminate type, determine if this value is supported by the plugin."""
        return isinstance(value, (Table, TableMetaclass))

    def from_pydantic_model_instance(self, model_class: Type[Table], pydantic_model_instance: "BaseModel") -> Table:
        """Given an instance of a pydantic model created using the plugin's ``to_pydantic_model_class``, return an
        instance of the class from which that pydantic model has been created.

        This class is passed in as the ``model_class`` kwarg.
        """
        return self.from_dict(model_class=model_class, **pydantic_model_instance.dict())

    def to_dict(self, model_instance: Table) -> Dict[str, Any]:
        """Given an instance of a model supported by the plugin, return a dictionary of serializable values."""
        return model_instance.to_dict()

    def from_dict(self, model_class: Type[Table], **kwargs: Any) -> Table:
        """Given a class supported by this plugin and a dict of values, create an instance of the class."""
        instance = model_class()
        for column in instance.all_columns():
            meta = column._meta
            if meta.name in kwargs:
                setattr(instance, meta.name, kwargs[meta.name])
        return instance



from typing import TYPE_CHECKING, Any, Dict, Literal, Optional, Union
from urllib.parse import quote

from starlite.constants import REDIRECT_STATUS_CODES
from starlite.enums import MediaType
from starlite.exceptions import ImproperlyConfiguredException
from starlite.response.base import Response
from starlite.status_codes import HTTP_307_TEMPORARY_REDIRECT

if TYPE_CHECKING:
    from starlite.datastructures import BackgroundTask, BackgroundTasks
    from starlite.types import ResponseCookies


class RedirectResponse(Response[Any]):
    """A redirect response."""

    def __init__(
        self,
        url: str,
        *,
        status_code: Literal[301, 302, 303, 307, 308] = HTTP_307_TEMPORARY_REDIRECT,
        background: Optional[Union["BackgroundTask", "BackgroundTasks"]] = None,
        headers: Optional[Dict[str, Any]] = None,
        cookies: Optional["ResponseCookies"] = None,
        encoding: str = "utf-8",
    ) -> None:
        """Initialize the response.

        Args:
            url: A url to redirect to.
            status_code: An HTTP status code. The status code should be one of 301, 302, 303, 307 or 308,
                otherwise an exception will be raised. .
            headers: A string keyed dictionary of response headers. Header keys are insensitive.
            cookies: A list of :class:`Cookie <starlite.datastructures.Cookie>` instances to be set under the response 'Set-Cookie' header.
            encoding: The encoding to be used for the response headers.

        Raises:
            :class:`ImproperlyConfiguredException <starlite.exceptions.ImproperlyConfiguredException>`: If status code is not a redirect status code.
        """
        if status_code not in REDIRECT_STATUS_CODES:
            raise ImproperlyConfiguredException(
                f"{status_code} is not a valid for this response. "
                f"Redirect responses should have one of "
                f"the following status codes: {', '.join([str(s) for s in REDIRECT_STATUS_CODES])}"
            )
        super().__init__(
            background=background,
            content=b"",
            cookies=cookies,
            headers={**(headers or {}), "location": quote(url, safe="/#%[]=:;$&()+,!?*@'~")},
            media_type=MediaType.TEXT,
            status_code=status_code,
            encoding=encoding,
        )



from typing import Any, Dict

from pydantic import validator
from pydantic_openapi_schema.v3_1_0 import Header


class ResponseHeader(Header):
    """Container type for a response header."""

    documentation_only: bool = False
    """Defines the ResponseHeader instance as for OpenAPI documentation purpose only."""
    value: Any = None
    """Value to set for the response header."""

    @validator("value", always=True)
    def validate_value(cls, value: Any, values: Dict[str, Any]) -> Any:  # pylint: disable=no-self-argument
        """Ensure that either value is set or the instance is for documentation_only."""
        if values.get("documentation_only") or value is not None:
            return value
        raise ValueError("value must be set if documentation_only is false")



from starlite.testing.client.async_client import AsyncTestClient
from starlite.testing.client.base import BaseTestClient
from starlite.testing.client.sync_client import TestClient
from starlite.testing.create_test_client import create_test_client
from starlite.testing.request_factory import RequestFactory
from starlite.testing.transport import (
    ConnectionUpgradeException,
    SendReceiveContext,
    TestClientTransport,
)

__all__ = (
    "TestClient",
    "RequestFactory",
    "create_test_client",
    "TestClientTransport",
    "AsyncTestClient",
    "BaseTestClient",
    "ConnectionUpgradeException",
    "SendReceiveContext",
)



from typing import Any, Protocol


class Logger(Protocol):  # pragma: no cover
    """Logger protocol."""

    def debug(self, event: str, *args: Any, **kwargs: Any) -> Any:
        """Output a log message at 'DEBUG' level.

        Args:
             event: Log message.
             *args: Any args.
             **kwargs: Any kwargs.
        """

    def info(self, event: str, *args: Any, **kwargs: Any) -> Any:
        """Output a log message at 'INFO' level.

        Args:
             event: Log message.
             *args: Any args.
             **kwargs: Any kwargs.
        """

    def warning(self, event: str, *args: Any, **kwargs: Any) -> Any:
        """Output a log message at 'WARNING' level.

        Args:
             event: Log message.
             *args: Any args.
             **kwargs: Any kwargs.
        """

    def warn(self, event: str, *args: Any, **kwargs: Any) -> Any:
        """Output a log message at 'WARN' level.

        Args:
             event: Log message.
             *args: Any args.
             **kwargs: Any kwargs.
        """

    def error(self, event: str, *args: Any, **kwargs: Any) -> Any:
        """Output a log message at 'ERROR' level.

        Args:
             event: Log message.
             *args: Any args.
             **kwargs: Any kwargs.
        """

    def fatal(self, event: str, *args: Any, **kwargs: Any) -> Any:
        """Output a log message at 'FATAL' level.

        Args:
             event: Log message.
             *args: Any args.
             **kwargs: Any kwargs.
        """

    def exception(self, event: str, *args: Any, **kwargs: Any) -> Any:
        """Log a message with level 'ERROR' on this logger. The arguments are interpreted as for debug(). Exception info
        is added to the logging message.

        Args:
             event: Log message.
             *args: Any args.
             **kwargs: Any kwargs.
        """

    def critical(self, event: str, *args: Any, **kwargs: Any) -> Any:
        """Output a log message at 'INFO' level.

        Args:
             event: Log message.
             *args: Any args.
             **kwargs: Any kwargs.
        """
